#ifndef _GNU_SOURCE
# define _GNU_SOURCE
#endif
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <queue>
#include <stack>
#include <map>
#include <algorithm>
#include <valarray>
#include <cstdlib>
#include <climits>
#include <cmath>
#include <cassert>
#include <cfloat>
#include <unistd.h>
#include <stdint.h>
#include <getopt.h>
#include "arith.h"
#include "ply.h"
#include "coders2d.h"
#include "ANN/ANN.h"

enum predictor_t
{
    PRED_PLAIN = 0,
    PRED_LEFT = 1,
    PRED_RIGHT = 2,
    PRED_FORWARD = 3,
    PRED_COUNT = 4
};

enum coder_t
{
    CODER_AXIAL,
    CODER_RADIAL,
    CODER_SINGLE
};

using namespace std;

static int slewn = -1;
static int slewd = 1;
static bool full_mst = true;
static int predictor_set = 0xf;
static coder_t coder_mode = CODER_SINGLE;
static string dump_file = "";
static string ps_file = "";
static double ps_zoom_from[4] = {DBL_MIN, DBL_MIN, DBL_MIN, DBL_MIN};
static double ps_zoom_to[4] = {DBL_MIN, DBL_MIN, DBL_MIN, DBL_MIN};
static bool compressing = true;

const char * const predictor_names[PRED_COUNT + 1] =
{
    "unbiased",
    "left",
    "right",
    "forward",
    "terminate"
};

const int predictor_vectors[PRED_COUNT][2] =
{
    {0, 0},
    {0, -1},
    {0, 1},
    {1, 0},
};

struct edge
{
    int target;

    edge() {}
    explicit edge(int t) : target(t) {}
};

struct flip_vector
{
    valarray<int> value;
    predictor_t predictor;

    flip_vector() : value(2), predictor(PRED_PLAIN) {}
};

struct node
{
    valarray<float> rcoord;
    valarray<int> coord;
    vector<edge> edges;
    int parent;
    int discovery;
    vector<int> children;
    double vertex_bits;
    flip_vector cooked;

    node() : rcoord(2), coord(2), edges(), parent(-1), discovery(-1),
        children(), vertex_bits(0.0), cooked() {}
};

struct cluster_data
{
    vector<node> nodes;
    vector<int> order;
    double mst_max;
    float scale;
    valarray<float> corner;
};

template<class T>
static ostream &operator <<(ostream &o, const valarray<T> d)
{
    o << d[0] << " " << d[1];
    return o;
}

template<class T, class S>
static T dot(const valarray<T> &x, const valarray<S> &y)
{
    return x[0] * y[0] + x[1] * y[1];
}

static inline double length(const valarray<int> &x)
{
    return sqrt((double) x[0] * (double) x[0]
                + (double) x[1] * (double) x[1]);
}

static inline float length(const valarray<float> &x)
{
    return sqrtf(x[0] * x[0] + x[1] * x[1]);
}

static inline int round_length(const valarray<int> &x)
{
    return (int) round(length(x));
}

static inline int ceil_length(const valarray<int> &x)
{
    return (int) ceil(length(x));
}

/* Unused in 2D version
static bool is_zero(const valarray<int> &x)
{
    return x.min() == 0 && x.max() == 0;
}

static bool is_zero(const valarray<float> &x)
{
    return length(x) == 0.0f;
}
*/

static int metric(const valarray<int> &x)
{
    return dot(x, x);
}

static inline valarray<int> get_delta(int x, const cluster_data &data)
{
    const vector<node> &nodes = data.nodes;
    if (nodes[x].parent < 0) return valarray<int>(0, 2);
    else return nodes[x].coord - nodes[data.nodes[x].parent].coord;
}

/* Returns a prediction using the given predictor, returning the
 * predicted point relative to the parent.
 */
static valarray<int> make_prediction(const valarray<int> &old_forward,
                                     predictor_t predictor)
{
    valarray<int> old_left(0, 2);
    old_left[0] = -old_forward[1];
    old_left[1] = old_forward[0];

    valarray<int> pred(3);
    valarray<int> p(3);

    for (int j = 0; j < 3; j++)
        pred[j] = predictor_vectors[predictor][0] * old_forward[j]
            + predictor_vectors[predictor][1] * old_left[j];
    return pred;
}

static flip_vector update(const valarray<int> &delta,
                          const valarray<int> &old_forward)
{
    flip_vector out;
    int best_metric = INT_MAX;

    for (int i = 0; i < (int) PRED_COUNT; i++)
    {
        valarray<int> p(0, 3);

        if (!(predictor_set & (1 << i))) continue;
        p = delta - make_prediction(old_forward, (predictor_t) i);
        if (metric(p) < best_metric)
        {
            out.value = p;
            out.predictor = (predictor_t) i;
            best_metric = metric(p);
        }
    }

    return out;
}

struct mst_data
{
    priority_queue<pair<float, int>, vector<pair<float, int> >, greater<pair<float, int> > > q;
    vector<bool> done;
    vector<float> prio;
    int run;
    bool fancy;
};

static float mst_metric(int cur, int nxt,
                        const mst_data &mst, const cluster_data &data)
{
    const vector<node> &nodes = data.nodes;
    valarray<int> delta = nodes[nxt].coord - nodes[cur].coord;

    if (slewn < 0 && nodes[cur].parent >= 0)
    {
        valarray<int> delta2 = nodes[cur].coord - nodes[nodes[cur].parent].coord;
        float m = length(delta2 - delta);
        float d = log(m + 1.0) * length(delta);
        return d;
    }
    else if (slewn != 0 && nodes[cur].parent >= 0)
    {
        valarray<int> delta2 = nodes[cur].coord - nodes[nodes[cur].parent].coord;
        float x = dot(delta, delta2) / length(delta2);
        float y = sqrt(max(0.0f, dot(delta, delta) - x * x));
        if (x < 0.0f) return length(delta);
        else return sqrtf(x * x * slewn / slewd + y * y);
    }
    else
        return length(delta);
}

static void mst_update(cluster_data &data, mst_data &mst, int cur)
{
    vector<node> &nodes = data.nodes;

    mst.done[cur] = true;
    nodes[cur].discovery = mst.run;
    data.order[mst.run++] = cur;
    valarray<int> delta = get_delta(cur, data);
    valarray<int> old(0, 2);
    if (nodes[cur].parent != -1)
        old = get_delta(nodes[cur].parent, data);
    nodes[cur].cooked = update(delta, old);

    for (size_t i = 0; i < nodes[cur].edges.size(); i++)
    {
        int nxt = nodes[cur].edges[i].target;
        float p;
        if (mst.done[nxt]) continue;

        p = mst.fancy ? mst_metric(cur, nxt, mst, data)
            : metric(nodes[nxt].coord - nodes[cur].coord);
        if (p < mst.prio[nxt])
        {
            mst.prio[nxt] = p;
            nodes[nxt].parent = cur;
            mst.q.push(make_pair(p, nxt));
        }
    }
}

static bool mst_mesh(cluster_data &data, bool fancy)
{
    vector<node> &nodes = data.nodes;
    size_t N = nodes.size();

    mst_data mst;
    mst.done.resize(N, false);
    mst.prio.resize(N, INT_MAX);
    mst.fancy = fancy;
    mst.run = 0;
    data.order.resize(N);

    mst.prio[0] = 0;
    mst.q.push(make_pair(0, 0));
    while (!mst.q.empty())
    {
        int cur = mst.q.top().second;
        mst.q.pop();
        if (mst.done[cur]) continue;
        mst_update(data, mst, cur);
        if (fancy && cur != 0)
        {
            valarray<int> last = get_delta(cur, data);
            do
            {
                int nxt = -1;
                float nxt_metric = FLT_MAX;
                for (size_t i = 0; i < nodes[cur].edges.size(); i++)
                {
                    int trg = nodes[cur].edges[i].target;
                    if (mst.done[trg]) continue;
                    float trg_metric = mst_metric(cur, trg, mst, data);
                    if (trg_metric < nxt_metric)
                    {
                        nxt = trg;
                        nxt_metric = trg_metric;
                    }
                }
                if (nxt != -1)
                {
                    nodes[nxt].parent = cur;
                    mst_update(data, mst, nxt);
                }
                cur = nxt;
            } while (cur != -1);
        }
    }
    return mst.run == (int) N;
}

static void create_edges(cluster_data &data, ANNpointSet *t, int high)
{
    vector<node> &nodes = data.nodes;
    size_t N = nodes.size();

    for (size_t i = 0; i < N; i++)
    {
        ANNcoord q[2];
        q[0] = nodes[i].coord[0];
        q[1] = nodes[i].coord[1];
        int k = t->annkFRSearch(q, high * high + 0.5);
        ANNidxArray nn_idx = new ANNidx[k];
        ANNdistArray dd = new ANNdist[k];
        t->annkFRSearch(q, high * high + 0.5, k, nn_idx, dd);

        nodes[i].edges.clear();
        for (int j = 0; j < k; j++)
        {
            int id = nn_idx[j];
            nodes[i].edges.push_back(edge(id));
        }

        delete[] nn_idx;
        delete[] dd;
    }
}

static bool connected(const cluster_data &data)
{
    const vector<node> &nodes = data.nodes;
    size_t N = nodes.size();
    vector<bool> done(N, false);
    vector<int> st;

    st.reserve(N);
    st.push_back(0);
    N--;
    while (!st.empty())
    {
        int cur = st.back();
        st.pop_back();
        done[cur] = true;
        for (size_t i = 0; i < nodes[cur].edges.size(); i++)
        {
            int trg = nodes[cur].edges[i].target;
            if (!done[trg])
            {
                done[trg] = true;
                st.push_back(trg);
                N--;
            }
        }
    }
    return N == 0;
}

static void mst_full(cluster_data &data)
{
    vector<node> &nodes = data.nodes;
    size_t N = nodes.size();
    ANNcoord *points = new ANNcoord[2 * N];
    ANNpoint *point_array = new ANNpoint[N];
    ANNpointSet *t;
    int lmax = 1;

    for (size_t i = 0; i < N; i++)
    {
        points[i * 2 + 0] = nodes[i].coord[0];
        points[i * 2 + 1] = nodes[i].coord[1];
        point_array[i] = points + i * 2;
    }
    t = new ANNkd_tree(point_array, N, 2);

    do
    {
        lmax *= 2;
        create_edges(data, t, lmax);
    } while (!connected(data));
    mst_mesh(data, false);

    data.mst_max = 0;
    for (size_t i = 0; i < N; i++)
    {
        valarray<int> delta = get_delta(i, data);
        data.mst_max = max(data.mst_max, sqrt((double) dot(delta, delta)));
    }
    lmax = (int) ceil(data.mst_max);
    create_edges(data, t, lmax);
    mst_mesh(data, true);

    delete t;
    delete[] points;
    delete[] point_array;
}

struct coder_data
{
    int max_length;
    int max_x;
    int max_y;
    coder *geometry;
    context predictor_context;

    coder_data() : max_length(0), max_x(0), max_y(0), geometry(NULL), predictor_context() {}
};

template<class T>
static void show_context(const char *name, const T &ctx)
{
    double itotal = 1.0 / ctx.use();
    cout << name << ": "
        << ctx.bits() * itotal << " | "
        << ctx.entropy() * itotal << " ("
        << ctx.use() << ")\n";
}

class predictor_comparator : public binary_function<int, int, bool>
{
private:
    const cluster_data &data;

public:
    predictor_comparator(const cluster_data &d) : data(d) {}

    bool operator()(int a, int b)
    {
        return data.nodes[a].cooked.predictor < data.nodes[b].cooked.predictor;
    }
};

/* Reorders vertices based on a depth first walk of the tree, visiting
 * children in the order determined by the predictor.
 */
static void reorder(cluster_data &data)
{
    vector<node> &nodes = data.nodes;
    stack<int> s;
    int index = 0;

    s.push(0);
    while (!s.empty())
    {
        int cur = s.top();
        s.pop();
        data.order[index++] = cur;
        sort(nodes[cur].children.begin(), nodes[cur].children.end(), predictor_comparator(data));
        for (vector<int>::reverse_iterator i = nodes[cur].children.rbegin();
             i != nodes[cur].children.rend(); i++)
            s.push(*i);
    }
}

static void cook(cluster_data &data, coder_data *coders)
{
    vector<node> &nodes = data.nodes;
    size_t N = nodes.size();

    for (size_t i = 1; i < N; i++)
    {
        nodes[nodes[i].parent].children.push_back(i);
        int c = nodes[i].cooked.predictor;
        coders[c].max_length = max(coders[c].max_length, ceil_length(nodes[i].cooked.value));
        coders[c].max_x = max(coders[c].max_x, abs(nodes[i].cooked.value[0]));
        coders[c].max_y = max(coders[c].max_y, abs(nodes[i].cooked.value[1]));
    }

    reorder(data);
}

static void prepare_encoders(coder_data *coders)
{
    for (int i = 0; i < (int) PRED_COUNT; i++)
    {
        switch (coder_mode)
        {
        case CODER_AXIAL:
            coders[i].geometry = new axial_coder(coders[i].max_x,
                                                 coders[i].max_y);
            break;
        case CODER_RADIAL:
            coders[i].geometry = new radial_coder(coders[i].max_length);
            break;
        case CODER_SINGLE:
            coders[i].geometry = new single_coder(coders[i].max_x,
                                                  coders[i].max_y);
            break;
        }
    }

    for (int i = 0; i < (int) PRED_COUNT; i++)
        coders[i].predictor_context.code_range(i, PRED_COUNT);
    coders[PRED_COUNT].predictor_context.code_range(0, PRED_COUNT);
}

static void compress(cluster_data &data, compressor &c, coder_data *coders)
{
    vector<node> &nodes = data.nodes;
    size_t N = nodes.size();

    for (size_t i = 0; i < N; i++)
    {
        int cur = data.order[i];

        // Encode position. The parent provides the predictor
        if (cur != 0)
        {
            int predictor = nodes[cur].cooked.predictor;
            nodes[cur].vertex_bits += coders[predictor].geometry->encode(c, nodes[cur].cooked.value);
        }

        // Encode valence and child predictors
        predictor_t last = PRED_COUNT;
        for (size_t j = 0; j < nodes[cur].children.size(); j++)
        {
            int nxt = nodes[cur].children[j];
            nodes[nxt].vertex_bits += c.encode(coders[last].predictor_context, nodes[nxt].cooked.predictor);
            last = nodes[nxt].cooked.predictor;
        }
        nodes[cur].vertex_bits += c.encode(coders[last].predictor_context, PRED_COUNT);
    }
}

static void predictor_stats(predictor_t predictor,
                            const vector<cluster_data> &clusters)
{
    size_t count = 0;
    double l2 = 0.0;

    for (size_t i = 0; i < clusters.size(); i++)
    {
        const vector<node> &nodes = clusters[i].nodes;
        size_t N = nodes.size();

        for (size_t j = 1; j < N; j++)
        {
            if (predictor == PRED_COUNT
                || predictor == nodes[j].cooked.predictor)
            {
                count++;
                l2 += dot(nodes[j].cooked.value, nodes[j].cooked.value);
            }
        }
    }
    cout << "    count: " << count << "\n";
    if (count)
        cout << "    L2: " << sqrt(l2 / count) << "\n";
    else
        cout << "    L2: -1\n";
}

static void statistics(compressor &c, const vector<cluster_data> &clusters, coder_data *coders)
{
    size_t points = 0;

    for (size_t i = 0; i < clusters.size(); i++)
        points += clusters[i].nodes.size();
    double context_bits = 0.0, context_entropy = 0.0;
    for (int i = 0; i <= (int) PRED_COUNT; i++)
    {
        string title = string("predictor (") + predictor_names[i] + ")";
        show_context(title.c_str(), coders[i].predictor_context);
        context_bits += coders[i].predictor_context.bits();
        context_entropy += coders[i].predictor_context.entropy();
    }
    for (int i = 0; i < (int) PRED_COUNT; i++)
    {
        cout << predictor_names[i] << "\n"
            << "    max length: " << coders[i].max_length << "\n";
        show_context("    geometry", *coders[i].geometry);
        predictor_stats((predictor_t) i, clusters);
    }
    cout << "total\n";
    predictor_stats(PRED_COUNT, clusters);
    cout << "points: " << points << "\n";
    cout << "tree: "
        << context_bits / points << " | "
        << context_entropy / points << "\n";
    cout << "geometry: "
        << (c.bits() - context_bits) / points << "\n";
    cout << "total: " << c.bits() / (double) points << "\n";
    cout << "\n";
    // coders[PRED_FORWARD].geometry->dump(cout);
}

static bool optarg_bool()
{
    if (!optarg) return true;
    else if (strcmp(optarg, "yes") == 0) return true;
    else if (strcmp(optarg, "no") == 0) return false;
    else
    {
        cerr << "Please specify 'yes' or 'no'\n";
        exit(1);
    }
}

static vector<string> optarg_commas()
{
    char *cur, *nxt;
    vector<string> ans;

    cur = optarg;
    while (*cur)
    {
        while (*cur == ',') cur++;
        if (*cur == '\0') break;
        nxt = cur;
        while (*nxt != ',' && *nxt != '\0') nxt++;
        ans.push_back(string(cur, nxt - cur));
        cur = nxt;
    }
    return ans;
}

static int optarg_predictor_set()
{
    vector<string> predictors = optarg_commas();
    int ps = 0;

    for (size_t i = 0; i < predictors.size(); i++)
    {
        int j;
        for (j = 0; j < PRED_COUNT; j++)
            if (predictors[i] == predictor_names[i]) break;
        if (j == PRED_COUNT)
        {
            cerr << "Please specify a set of predictors, separated by commas\n";
            exit(1);
        }
        ps |= 1 << j;
    }
    return ps;
}

static void optarg_zoom(double *zoom)
{
    vector<string> fields = optarg_commas();

    if (fields.size() != 4)
    {
        cerr << "Please specify zoom location in form x1,y1,x2,y2\n";
        exit(1);
    }
    zoom[0] = atof(fields[0].c_str());
    zoom[1] = atof(fields[1].c_str());
    zoom[2] = atof(fields[2].c_str());
    zoom[3] = atof(fields[3].c_str());
}

static void process_args(int *argc, char ***argv)
{
    int c;
    enum longopt_t
    {
        ZOOM_FROM = 256,
        ZOOM_TO
    };
    const struct option longopts[] =
    {
        { "dump", 1, NULL, 'd' },
        { "predictors", 1, NULL, 'p' },
        { "mst", 1, NULL, 'm' },
        { "slewn", 1, NULL, 's' },
        { "slewd", 1, NULL, 'd' },
        { "encoder", 1, NULL, 'e' },
        { "ps", 1, NULL, 'P' },
        { "decompress", 1, NULL, 'D' },
        { "zoom-from", 1, NULL, (int) ZOOM_FROM },
        { "zoom-to", 1, NULL, (int) ZOOM_TO },
        { NULL, 0, NULL, 0 }
    };

    while ((c = getopt_long(*argc, *argv, "d:p:m:s:S:e:P:D", longopts, NULL)) != -1)
        switch (c)
        {
        case 'd':
            dump_file = optarg;
            break;
        case 'P':
            ps_file = optarg;
            break;
        case 'p':
            predictor_set = optarg_predictor_set();
            break;
        case 'm':
            full_mst = optarg_bool();
            break;
        case 's':
            slewn = atoi(optarg);
            break;
        case 'S':
            slewd = atoi(optarg);
            break;
        case 'e':
            if (strcmp(optarg, "radial") == 0)
                coder_mode = CODER_RADIAL;
            else if (strcmp(optarg, "axial") == 0)
                coder_mode = CODER_AXIAL;
            else if (strcmp(optarg, "single") == 0)
                coder_mode = CODER_SINGLE;
            else
            {
                cerr << "No such encoder " << optarg << "\n";
                exit(1);
            }
            break;
        case 'D': compressing = false; break;
        case ZOOM_FROM:
            optarg_zoom(ps_zoom_from);
            break;
        case ZOOM_TO:
            optarg_zoom(ps_zoom_to);
            break;
        case ':':
        case '?':
            exit(0);
        }
    *argc -= optind;
    *argv += optind;
}

static void read_cluster(istream &in, double quant,
                         vector<node> &vertices)
{
    int N;
    string line;

    in >> ws;
    getline(in, line);   // origin
    getline(in, line);   // normal
    getline(in, line);   // bound1
    getline(in, line);   // bound2
    in >> N;

    vertices.resize(N);
    for (int i = 0; i < N; i++)
    {
        double z;
        node &cur = vertices[i];

        in >> cur.rcoord[0] >> cur.rcoord[1] >> z;
        cur.coord[0] = (int) rint(cur.rcoord[0] / quant);
        cur.coord[1] = (int) rint(cur.rcoord[1] / quant);
    }
}

static void readin(const char *fname, vector<cluster_data> &clusters)
{
    int C;
    double quant;

    ifstream f(fname);
    if (!f)
    {
        cerr << "Cannot open " << fname << "\n";
        exit(1);
    }
    f >> C >> quant;
    clusters.resize(C);
    for (int c = 0; c < C; c++)
        read_cluster(f, quant, clusters[c].nodes);
}

static void quantise(int bits, cluster_data &data)
{
    size_t bottom = 0;
    for (size_t i = 1; i < data.nodes.size(); i++)
    {
        bool smaller = false;
        for (int j = 2; j >= 0; j--)
            if (data.nodes[i].coord[j] != data.nodes[bottom].coord[j])
            {
                smaller = data.nodes[i].coord[j] < data.nodes[bottom].coord[j];
                break;
            }
        if (smaller) bottom = i;
    }

    swap(data.nodes[0], data.nodes[bottom]);
    for (size_t i = 0; i < data.nodes.size(); i++)
        for (size_t j = 0; j < data.nodes[i].edges.size(); j++)
        {
            if (data.nodes[i].edges[j].target == 0)
                data.nodes[i].edges[j].target = bottom;
            else if (data.nodes[i].edges[j].target == (int) bottom)
                data.nodes[i].edges[j].target = 0;
        }
}

static void write_header(ostream &out, const vector<cluster_data> &clusters,
                         const coder_data *coders)
{
    uint32_t C = clusters.size();
    out.write((const char *) &C, sizeof(C));

    for (uint32_t c = 0; c < C; c++)
    {
        const cluster_data &data = clusters[c];
        uint32_t N = data.nodes.size();
        uint8_t ps = predictor_set;
        uint8_t cm = coder_mode;
        out.write((const char *) &N, sizeof(N));
        out.write((const char *) &data.scale, sizeof(data.scale));
        for (int i = 0; i < 2; i++)
            out.write((const char *) &data.corner[i], sizeof(data.corner[i]));
        for (int i = 0; i < 2; i++)
        {
            uint32_t w = data.nodes[0].coord[i];
            out.write((const char *) &w, sizeof(w));
        }
        out.write((const char *) &ps, sizeof(ps));
        out.write((const char *) &cm, sizeof(cm));
        for (int i = 0; i < PRED_COUNT; i++)
            coders[i].geometry->write_header(out);
    }
}

/**********************
 * Decompression code *
 **********************/

static void read_header(istream &in, vector<cluster_data> &clusters,
                        coder_data *coders)
{
    uint32_t C, N;
    uint8_t ps;
    uint8_t cm;
    uint32_t w;

    in.read((char *) &C, sizeof(C));
    clusters.resize(C);
    for (uint32_t c = 0; c < C; c++)
    {
        cluster_data &data = clusters[c];
        in.read((char *) &N, sizeof(N));
        in.read((char *) &data.scale, sizeof(data.scale));
        data.corner.resize(2);
        for (int i = 0; i < 2; i++)
            in.read((char *) &data.corner[i], sizeof(data.corner[i]));
        data.nodes.resize(N);
        for (int i = 0; i < 2; i++)
        {
            in.read((char *) &w, sizeof(w));
            data.nodes[0].coord[i] = w;
        }
        in.read((char *) &ps, sizeof(ps)); predictor_set = ps;
        in.read((char *) &cm, sizeof(cm)); coder_mode = (coder_t) cm;
        for (int i = 0; i < PRED_COUNT; i++)
        {
            switch (cm)
            {
            case CODER_AXIAL:
                coders[i].geometry = new axial_coder(in); break;
            case CODER_RADIAL:
                coders[i].geometry = new radial_coder(in); break;
            case CODER_SINGLE:
                coders[i].geometry = new single_coder(in); break;
            default: abort();
            }
            coders[i].predictor_context.code_range(i, PRED_COUNT);
        }
        coders[PRED_COUNT].predictor_context.code_range(0, PRED_COUNT);
        data.nodes.resize(N);
    }
}

static void decompress(decompressor &d, cluster_data &data,
                       coder_data *coders)
{
    vector<node> &nodes = data.nodes;
    size_t N = 1;
    stack<int> s;

    s.push(0);
    while (!s.empty())
    {
        int cur = s.top();
        s.pop();
        if (cur != 0)
        {
            /* Recover position. The parent provided the predictor */

            int parent = nodes[cur].parent;
            predictor_t predictor = nodes[cur].cooked.predictor;
            valarray<int> forward = get_delta(parent, data);
            valarray<int> base = make_prediction(forward,
                                                 predictor);
            valarray<int> correction = coders[predictor].geometry->decode(d);
            nodes[cur].coord = nodes[parent].coord + base + correction;
        }

        /* Recover predictors and valence */
        predictor_t last = PRED_COUNT;
        while (1)
        {
            predictor_t predictor = (predictor_t) d.decode(coders[last].predictor_context);
            if (predictor == PRED_COUNT) break; /* terminator */
            nodes[N].cooked.predictor = predictor;
            nodes[cur].children.push_back(N);
            nodes[N].parent = cur;
            N++;
            last = predictor;
        }
        for (vector<int>::reverse_iterator i = nodes[cur].children.rbegin();
             i != nodes[cur].children.rend(); i++)
            s.push(*i);
    }
}

static void dequantise(cluster_data &data)
{
    vector<node> &nodes = data.nodes;
    size_t N = nodes.size();

    for (size_t i = 0; i < N; i++)
        for (int j = 0; j < 2; j++)
            nodes[i].rcoord[j] = data.corner[j] + nodes[i].coord[j] / data.scale;
}

static void writeout(const char *fname, const vector<cluster_data> &data)
{
}

int main(int argc, char **argv)
{
    coder_data coders[PRED_COUNT + 1];
    vector<cluster_data> clusters;

    process_args(&argc, &argv);

    if (compressing)
    {
        assert(argc == 3);

        compressor c;
        ofstream out(argv[2]);
        out.exceptions(ios::failbit);

        readin(argv[0], clusters);
        for (size_t i = 0; i < clusters.size(); i++)
        {
            quantise(atoi(argv[1]), clusters[i]);
            if (full_mst) mst_full(clusters[i]);
            else mst_mesh(clusters[i], true);
            cook(clusters[i], coders);
        }
        prepare_encoders(coders);
        write_header(out, clusters, coders);
        c.open(out);
        for (size_t i = 0; i < clusters.size(); i++)
        {
            compress(clusters[i], c, coders);
            /* No 2D dump file format yet! */
            /* if (dump_file != "") dump(dump_file, data); */
            /* if (ps_file != "") dump_ps(ps_file, clusters[i]); */
            cout << "L: " << clusters[i].mst_max << "\n";
            /* Need to split coder stats from cluster stats */
        }

        statistics(c, clusters, coders);
        c.close();
        out.close();
    }
    else
    {
        assert(argc == 2);
        ifstream in(argv[0]);
        in.exceptions(ios::failbit);

        read_header(in, clusters, coders);
        decompressor d(in);
        for (size_t i = 0; i < clusters.size(); i++)
        {
            decompress(d, clusters[i], coders);
            dequantise(clusters[i]);
        }
        in.close();
        writeout(argv[1], clusters);
    }
    return 0;
}
